from functools import partial
import time

class semistaticmethod(object):
    def __init__(self, callable):
        self.f = callable
    def __get__(self, obj, type=None):
        if (obj is None) and (type is not None):
            return partial(self.f, type)
        if (obj is not None):
            return partial(self.f, obj)
        return self.f
    @property
    def __func__(self):
        return self.f
    
def cls_method_timer(func):
    def wrapper(self, *args, **kwargs):
        t1 = time.time()
        result = func(self, *args, **kwargs)
        t2 = time.time()
        method_name = f"{type(self).__name__}::{func.__name__}"
        self.stdout.info(f'INFO: Task {method_name!r} executed in {(t2-t1):.4f}s')
        return result
    return wrapper 


from typing import Optional, Union, Dict, List

import matplotlib.patches as patches
import matplotlib.lines as lines
import numpy as np
import pandas as pd

from quickstats.plots.template import single_frame, parse_styles, format_axis_ticks, create_transform
from quickstats.plots import AbstractPlot
from quickstats.utils.common_utils import combine_dict

class UpperLimitBenchmarkPlot(AbstractPlot):
    
    STYLES = {
        'figure':{
            'figsize': (11.111, 8.333),
            'dpi': 72,
            'facecolor': "#FFFFFF"
        },
        'text':{
            'fontsize': 25
        },
        'ylabel':{
            'labelpad': 0
        }
    }
    
    COLOR_PALLETE = {
        '2sigma': 'hh:darkyellow',
        '1sigma': 'hh:lightturquoise',
        'expected': 'k',
        'observed': 'k',
    }
    
    LABELS = {
        '2sigma': r'Expected limit $\pm 2\sigma$',
        '1sigma': r'Expected limit $\pm 1\sigma$',
        'expected': r'Expected limit (95% CL)',
        'observed': r'Observed limit (95% CL)',
    }

    CONFIG = {     
        'xmargin': 0.3,
        'sigma_width': 0.6,
        'markersize': None
    }
    
    def __init__(self, benchmark_df, label_map,
                 limit_scale:float=1,
                 color_pallete:Optional[Dict]=None,
                 labels:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]='limit_benchmark',
                 config:Optional[Dict]=None,
                 analysis_label_options:Optional[Union[Dict, str]]=None):
        """
        Args:
            benchmar_df: pandas.DataFrame
                dataframe with columns ("-2", "-1", "0", "1", "2", "obs") representing
                the corresponding limit level and rows indexed by the benchmark names
        """
        super().__init__(color_pallete=color_pallete,
                         styles=styles, config=config,
                         analysis_label_options=analysis_label_options)
        self.benchmark_df = benchmark_df
        self.label_map = label_map
        self.limit_scale = limit_scale
        self.labels = combine_dict(self.LABELS, labels)
        self.hline_data = []
        
    def get_default_legend_order(self):
        return ['observed', 'expected', '1sigma', '2sigma']
    
    def add_hline(self, y:float,
                  label:str="Theory prediction",
                  **styles):
        hline_data = {
            'y'     : y,
            'label' : label,
            'styles': styles
        }
        self.hline_data.append(hline_data)
        self.legend_order.append(label)
        
    def draw_hlines(self, ax):
        for data in self.hline_data:
            h = ax.axhline(data['y'], label=data['label'], **data['styles'])
            self.update_legend_handles({data['label']: h})
    
    def draw(self, logy:bool=False, xlabel:str="", ylabel:str="", ylim=None,
             draw_observed:bool=True):
        
        eps = self.config['sigma_width'] / 2
        scale = self.limit_scale
        xmargin = self.config['xmargin']
        n_benchmarks = len(self.benchmark_df.index)
        
        ax = self.draw_frame(logy=logy)
        xlim = (0 - eps - xmargin, n_benchmarks - 1 + eps + xmargin)
        self.draw_axis_components(ax, ylabel=ylabel, xlabel=xlabel, ylim=ylim, xlim=xlim)
        
        xticklabels = []
        for index, (benchmark, limit) in enumerate(self.benchmark_df.iterrows()):
            benchmark_label = self.label_map.get(benchmark, benchmark)
            xticklabels.append(benchmark_label)
            h_2sigma = ax.fill_between([index - eps, index + eps], limit['-2'] * scale, limit['2'] * scale,
                                       facecolor=self.color_pallete['2sigma'],
                                       label=self.labels['2sigma'])
            h_1sigma = ax.fill_between([index - eps, index + eps], limit['-1'] * scale, limit['1'] * scale,
                                       facecolor=self.color_pallete['1sigma'],
                                       label=self.labels['1sigma'])
            h_exp = ax.scatter(x=index, y=limit['0'] * scale,
                               edgecolors=self.color_pallete['expected'],
                               facecolors='none', s=self.config['markersize'],
                               label=self.labels['expected'])
            self.update_legend_handles({"2sigma": h_2sigma, "1sigma": h_1sigma, "expected": h_exp})
            if draw_observed:
                h_obs = ax.scatter(x=index, y=limit['obs'] * scale,
                                   color=self.color_pallete['observed'],
                                   s=self.config['markersize'],
                                   label=self.labels['observed'])
                self.update_legend_handles({"observed": h_obs})
                
        self.draw_hlines(ax)
        
        ax.set_xticks(range(n_benchmarks))
        ax.set_xticklabels(xticklabels)
        
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        return ax


from typing import Dict, Optional, Union, List
import pandas as pd

from quickstats.plots import AbstractPlot
from quickstats.plots.template import create_transform
from quickstats.utils.common_utils import combine_dict

class Likelihood1DPlot(AbstractPlot):
    
    CONFIG = {
        'sigma_values': (1, 4),
        'sigma_pos': 0.93,
        'sigma_names': ('1 $\sigma$', '2 $\sigma$'),
        'sigma_line_styles':{
            'color': 'gray',
            'linestyle': '--'
        }
    }
    
    def __init__(self, data_map:Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 label_map:Optional[Dict]=None,
                 styles_map:Optional[Dict]=None,
                 color_cycle=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 config:Optional[Dict]=None):
        
        self.data_map = data_map
        self.label_map = label_map
        self.styles_map = styles_map
        
        super().__init__(color_cycle=color_cycle,
                         styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
        
    def get_default_legend_order(self):
        if not isinstance(self.data_map, dict):
            return []
        else:
            return list(self.data_map)
        
    def draw_single_data(self, ax, data:pd.DataFrame, 
                         styles:Optional[Dict]=None,
                         label:Optional[str]=None,
                         xattrib:str='mu', yattrib:str='qmu'):
        x = data[xattrib]
        y = data[yattrib]
        draw_styles = combine_dict(self.styles['plot'], styles)
        handle = ax.plot(x, y, **draw_styles, label=label)
        return handle[0]
    
    def draw(self, xattrib:str='mu', yattrib:str='qmu', xlabel:Optional[str]=None, 
             ylabel:Optional[str]="$-2\Delta ln(L)$",
             ymax:float=7, xmin:Optional[float]=None, xmax:Optional[float]=None,
             draw_sigma_line:bool=True, draw_sm_line:bool=False):
        ax = self.draw_frame()
        if isinstance(self.data_map, pd.DataFrame):
            self.draw_single_data(ax, self.data_map, self.styles_map,
                                  xattrib=xattrib, yattrib=yattrib)
        elif isinstance(self.data_map, dict):
            if self.styles_map is None:
                styles_map = {k:None for k in self.data_map}
            else:
                styles_map = self.styles_map
            if self.label_map is None:
                label_map = {k:k for k in data_map}
            else:
                label_map = self.label_map
            handles = {}
            for key in self.data_map:
                data = self.data_map[key]
                styles = styles_map.get(key, None)
                label = label_map.get(key, "")
                handle = self.draw_single_data(ax, data, styles, 
                                               label=label,
                                               xattrib=xattrib,
                                               yattrib=yattrib)
                handles[key] = handle
            self.update_legend_handles(handles)
        else:
            raise ValueError("invalid data format")
            
        if draw_sigma_line:
            transform = create_transform(transform_x="axis", transform_y="data")
            sigma_line_styles = self.config['sigma_line_styles']
            sigma_values = self.config['sigma_values']
            sigma_names = self.config['sigma_names']
            ax.hlines(sigma_values, xmin=0, xmax=1, zorder=0, transform=transform,
                      **sigma_line_styles)
            if sigma_names:
                sigma_pos = self.config['sigma_pos']
                for sigma_value, sigma_name in zip(sigma_values, sigma_names):
                    ax.text(sigma_pos, sigma_value, sigma_name, color='gray', ha='left',
                            va='bottom' if (sigma_pos > 0 and sigma_pos < 1) else 'center', fontsize=20, transform=transform)
            
        if draw_sm_line:
            transform = create_transform(transform_y="axis", transform_x="data")
            sm_line_styles = self.config['sm_line_styles']
            sm_values = self.config['sm_values']
            sm_names = self.config['sm_names']
            ax.vlines(sm_values, ymin=0, ymax=1, zorder=0, transform=transform,
                      **sm_line_styles)
            if sm_names:
                sm_pos = self.config['sm_pos']
                for sm_value, sm_name in zip(sm_values, sm_names):
                    ax.text(sm_value, sm_pos, sm_name, color='gray', ha='right', rotation=90,
                            va='bottom' if (sm_pos > 0 and sm_pos < 1) else 'center', fontsize=20, transform=transform)
            
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        ax.set_ylim([0, ymax])
        ax.set_xlim([xmin, xmax])

        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        
        return ax



import os
import json
import click

@click.command(name='likelihood_fit')
@click.option('-i', '--input_file', "filename", required=True, 
              help='Path to the input workspace file.')
@click.option('-o', '--outname', default='fit_result.json', show_default=True,
              help='Name of output.')
@click.option('--display/--no-display', default=True, show_default=True,
              help='Display fit result.')
@click.option('--save/--no-save', "save_result", default=False, show_default=True,
              help='Save fit result')
@click.option('--save_log/--skip_log', default=False, show_default=True,
              help='Save log file')
@click.option('--outdir', default="pulls", show_default=True,
              help='Output directory for pulls output')
@click.option('--export_as_np_pulls/--skip_export_as_np_pulls', default=False, show_default=True,
              help='Export (constrained) NP results for pulls plot')
@click.option('-w', '--workspace', 'ws_name', default=None, show_default=True,
              help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, show_default=True,
              help='Name of model config. Auto-detect by default.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of dataset.')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, show_default=True,
              help='Name of initial snapshot')
@click.option('-r', '--profile', 'profile_param', default="", show_default=True,
              help='Parameters to profile')
@click.option('-f', '--fix', 'fix_param', default="", show_default=True,
              help='Parameters to fix')
@click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, show_default=True,
              help='Use constrained NLL (i.e. include systematics)')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('-v', '--verbosity', default='INFO', show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
              help='Verbosity level.')
def likelihood_fit(**kwargs):
    """
    Perform likelihood fit on a workspace
    """
    _kwargs = {}
    for arg_name in ["outname", "save_log", "display", "save_result",
                     "export_as_np_pulls", "outdir"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _init_kwargs = {}
    for arg_name in ["filename", "data_name", "verbosity"]:
        _init_kwargs[arg_name] = kwargs.pop(arg_name)
    _init_kwargs['config'] = kwargs
    _init_kwargs['poi_name'] = []
    from quickstats.components import AnalysisBase   
    if _kwargs['save_log']:
        from quickstats.concurrent.logging import standard_log
        log_path = os.path.splitext(_kwargs["outname"])[0] + ".log"
        with standard_log(log_path) as logger:
            analysis = AnalysisBase(**_init_kwargs)
            fit_result = analysis.nll_fit(mode=3)
        print(f"INFO: Saved fit log to `{log_path}`")
    else:
        analysis = AnalysisBase(**_init_kwargs)
        fit_result = analysis.nll_fit(mode=3)
    output = {}
    output['fit_result'] = fit_result
    df = {'pois':{}, 'nuisance_parameters':{}}
    analysis.load_snapshot("currentSnapshot")
    df['pois']['prefit'] = analysis.model.as_dataframe('pois')
    df['nuisance_parameters']['prefit'] = analysis.model.as_dataframe('nuisance_parameters')
    analysis.load_snapshot("nllFit")
    df['pois']['postfit'] = analysis.model.as_dataframe('pois')
    df['nuisance_parameters']['postfit'] = analysis.model.as_dataframe('nuisance_parameters')
    if _kwargs['display']:
        import pandas as pd
        pd.set_option('display.max_rows', None)
    for key in ['pois', 'nuisance_parameters']:
        df[key]['combined'] = df[key]['prefit'].drop(["value", "error"], axis=1)
        df[key]['combined']['value_prefit'] = df[key]['prefit']['value']
        df[key]['combined']['value_postfit'] = df[key]['postfit']['value']
        df[key]['combined']['error_prefit'] = df[key]['prefit']['error']
        df[key]['combined']['error_postfit'] = df[key]['postfit']['error']
        output[key] = df[key]['combined'].to_dict("list")
        if _kwargs['display']:
            print("{}:".format(key.title()))
            print(df[key]['combined'])
            print()
    if _kwargs['save_result']:
        import json
        with open(_kwargs["outname"], "w") as f:
            json.dump(output, f, indent=2)
        print(f"INFO: Saved fit result to `{_kwargs['outname']}`")
    if _kwargs['export_as_np_pulls']:
        outdir = _kwargs['outdir']
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        nuis_df = df[key]['combined'].drop(['min', 'max', 'is_constant', 'error_prefit'], axis=1)
        nuis_df = nuis_df.rename(columns={"value_prefit":"nuis_nom", "name":"nuisance", 
                                          "value_postfit":"nuis_hat", "error_postfit":"nuis_hi"})
        nuis_df["nuis_lo"] = nuis_df["nuis_hi"]
        nuis_df["nuis_prefit"] = 1.0
        nuis_df = nuis_df.set_index(['nuisance'])
        constrained_np = [i.GetName() for i in analysis.model.get_constrained_nuisance_parameters()]
        nuis_df = nuis_df.loc[constrained_np].reset_index()
        nuis_data = nuis_df.to_dict('index')
        import json
        for i in nuis_data:
            data = nuis_data[i]
            np_name = data['nuisance']
            outpath = os.path.join(outdir, f"{np_name}.json")
            with open(outpath, "w") as outfile:
                json.dump({"nuis": data}, outfile, indent=2)
                
@click.command(name='np_correlation')
@click.option('-i', '--input_file', "filename", required=True, 
              help='Path to the input workspace file.')
@click.option('-o', '--basename', default='NP_correlation_matrix', show_default=True,
              help='Base name of the output.')
@click.option('--select', default=None, show_default=True,
              help='Select specific NPs to be stored in the final output (for json and plot only). '
                   'Use comma to separate the selection (wild card is supported).')
@click.option('--remove', default=None, show_default=True,
              help='Select specific NPs to be removed in the final output (for json and plot only). '
                   'Use comma to separate the selection (wild card is supported).')
@click.option('--save_plot/--no_save_plot', default=True, show_default=True,
              help='Save NP correlation matrix as a plot in pdf format')
@click.option('--save_json/--no_save_json', default=False, show_default=True,
              help='Save NP correlation matrix as a json file')
@click.option('--save_root/--no_save_root', default=False, show_default=True,
              help='Save NP correlation matrix as a 2D histogram in a root file')
@click.option('--plot_style', default="default", show_default=True,
              help='Plot style if save_plot is enabled. Choose between \"default\" and '
                   f'\"viridis\". Alternatively, a path to a yaml config file can be used')
@click.option('-w', '--workspace', 'ws_name', default=None, show_default=True,
              help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, show_default=True,
              help='Name of model config. Auto-detect by default.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of dataset.')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, show_default=True,
              help='Name of initial snapshot')
@click.option('-r', '--profile', 'profile_param', default="", show_default=True,
              help='Parameters to profile')
@click.option('-f', '--fix', 'fix_param', default="", show_default=True,
              help='Parameters to fix')
@click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, show_default=True,
              help='Use constrained NLL (i.e. include systematics)')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('-v', '--verbosity', default='INFO', show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
              help='Verbosity level.')
def np_correlation(**kwargs):
    """
    Evaluate post-fit NP correlation matrix
    """
    _kwargs = {}
    for arg_name in ["basename", "save_plot", "save_json", "save_root",
                     "plot_style", "select", "remove"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _init_kwargs = {}
    for arg_name in ["filename", "data_name", "verbosity"]:
        _init_kwargs[arg_name] = kwargs.pop(arg_name)
    _init_kwargs['config'] = kwargs
    _init_kwargs['poi_name'] = []
    from quickstats.components import AnalysisBase   
    analysis = AnalysisBase(**_init_kwargs)
    analysis.nll_fit(mode=3)
    fit_result = analysis.roofit_result
    basename = os.path.splitext(_kwargs['basename'])[0]
    from quickstats.utils.roofit_utils import get_correlation_matrix
    if _kwargs['save_root']:
        correlation_hist = get_correlation_matrix(fit_result, lib="root")
        outname = basename + ".root"
        correlation_hist.SaveAs(outname)
        print(f"INFO: Saved correlation histogram to `{outname}`")
        correlation_hist.Delete()
    from quickstats.utils.common_utils import filter_by_wildcards
    if _kwargs['save_json'] or _kwargs['save_plot']:
        df = get_correlation_matrix(fit_result, lib='pandas')
        labels = list(df.columns)
        selected = filter_by_wildcards(labels, _kwargs['select'])
        selected = filter_by_wildcards(selected, _kwargs['remove'], exclusion=True)
        to_drop = list(set(labels) - set(selected))
        df = df.drop(to_drop, axis=0).drop(to_drop, axis=1)
        if _kwargs['save_json']:
            data = df.to_dict()
            outname = basename + ".json"
            with open(outname, "w") as out:
                json.dump(data, out, indent=2)
            print(f"INFO: Saved correlation data to `{outname}`")
        if _kwargs['save_plot']:
            import matplotlib.pyplot as plt
            from quickstats.plots import CorrelationPlot
            plotter = CorrelationPlot(df)
            ax = plotter.draw_style(_kwargs['plot_style'])
            outname = basename + ".pdf"
            plt.savefig(outname, bbox_inches="tight")
            print(f"INFO: Saved correlation plot to `{outname}`")



class TObject:
    
    FUNDAMENTAL_TYPE = None
    
    @property
    def obj(self):
        return self._obj
    
    @obj.setter
    def obj(self, ref):
        if ref is None:
            self._obj = None
        elif not isinstance(ref, self.FUNDAMENTAL_TYPE):
            raise ValueError(f"Invalid ROOT object. Object must be an instance of {self.FUNDAMENTAL_TYPE}.")
        self._obj = ref
    
    def __init__(self, obj=None):
        self.obj = obj


from typing import Optional, Union, List, Dict

import ROOT

class Sample:
    
    def __init__(self):
        self.process = ""
        self.input_file = ""
        self.norm_factors = []
        self.shape_factors = []
        self.syst_groups = []
        self.expected = ROOT.RooArgSet()
        self.model_name = ""
        self.norm_name = ""
        self.share_pdf_group = ""
        
    def is_equal(self, other:Union[str, "Sample"]):
        if isinstance(other, str):
            return self.process == other
        elif isinstance(other, type(self)):
            return self.process == other.process
        else:
            raise TypeError(f"cannot compare object of type {type(other)} with "
                            "object of type Sample")
    def get_tag_name(self) -> str:
        is_shared_pdf = self.share_pdf_group != ""
        if is_shared_pdf:
            tag_name = self.share_pdf_group
        else:
            tag_name = self.process
        return tag_name


import cppyy
import ctypes
import numpy as np

from quickstats.interface import cppyy as interface_cppyy

ctype_maps = {
    np.dtype('bool'): ctypes.c_bool,
    np.dtype('byte'): ctypes.c_byte,
    np.dtype('ubyte'): ctypes.c_ubyte,
    np.dtype('short'): ctypes.c_short,
    np.dtype('ushort'): ctypes.c_ushort,
    np.dtype('intc'): ctypes.c_int,
    np.dtype('uintc'): ctypes.c_uint,
    np.dtype('single'): ctypes.c_float,
    np.dtype('float32'): ctypes.c_float,
    np.dtype('double'): ctypes.c_double,
    np.dtype('float64'): ctypes.c_double,
    np.dtype('int64'): ctypes.c_int64,
    np.dtype('int32'): ctypes.c_int32
}

ctype_str_maps = {
    ctypes.c_int : "int",
    ctypes.c_uint : "unsigned int",
    ctypes.c_float : "float",
    ctypes.c_double : "double",
    ctypes.c_bool : "bool",
    ctypes.c_byte : "byte",
    ctypes.c_ubyte : "unsigned char",
    ctypes.c_short : "short",
    ctypes.c_ushort : "unsigned short",
    ctypes.c_int64 : "long",
    ctypes.c_long : "long",
}

np_type_str_maps = {
    np.dtype('bool'): "bool",
    np.dtype('byte'): "char",
    np.dtype('ubyte'): "unsigned char",
    np.dtype('short'): "short",
    np.dtype('ushort'): "unsigned short",
    np.dtype('intc'): "int",
    np.dtype('uintc'): "unsigned int",
    np.dtype('single'): "float",
    np.dtype('float32'): "float",
    np.dtype('double'): "double",
    np.dtype('float64'): "double",
    np.dtype('int64'): "long long",
    np.dtype('int32'): "int"
}

typecode_to_np_type_maps = {
    'b': np.dtype('byte'),
    'B': np.dtype('ubyte'),
    'h': np.dtype('short'),
    'H': np.dtype('ushort'),
    'i': np.dtype('intc'),
    'I': np.dtype('uintc'),
    'l': np.dtype('int32'),
    'L': np.dtype('uint'),
    'q': np.dtype('int64'),
    'Q': np.dtype('ulonglong'),
    'f': np.dtype('float32'),
    'd': np.dtype('float64')
}

def as_vector(data:np.ndarray):
    if data.ndim != 1:
        raise ValueError("data must be 1D array")
    c_type = ctype_maps.get(data.dtype, None)
    if c_type is None:
        raise ValueError(f"unsupported data type \"{data.dtype}\"")
    c_type_p = ctypes.POINTER(c_type)
    c_type_str = ctype_str_maps[c_type]
    c_data = data.ctypes.data_as(c_type_p)
    size = data.shape[0]
    result = cppyy.gbl.VecUtils.as_vector[c_type_str](c_data, size)
    return result

def as_c_array(data:np.ndarray):
    if data.ndim != 1:
        raise ValueError("data must be 1D array")
    c_type = ctype_maps.get(data.dtype, None)
    if c_type is None:
        raise ValueError(f"unsupported data type \"{data.dtype}\"")
    c_type_p = ctypes.POINTER(c_type)
    c_type_str = ctype_str_maps[c_type]
    c_data = data.ctypes.data_as(c_type_p)
    return c_data

def as_np_array(vec:cppyy.gbl.std.vector):
    if vec.value_type == 'std::string':
        return np.array([str(v) for v in vec])
    return np.array(vec.data())

def c_array_to_np_array(arr:"cppyy.LowLevelView", size:int, shape=None, copy:bool=True):
    if not hasattr(arr, 'typecode'):
        raise RuntimeError(f"input must be an instance of cppyy.LowLevelView")
    np_type = typecode_to_np_type_maps.get(arr.typecode, None)
    if not np_type:
        raise RuntimeError(f"unsupported typecode \"{arr.typecode}\"")
    arr.reshape((size,))
    np_arr = np.frombuffer(arr, dtype=np_type, count=size)
    if shape is not None:
        return np_arr.reshape(shape)
    if copy:
        return np.array(np_arr)
    return np_arr


from typing import Optional, Dict, List
import uuid

def template_DSCB(observable:"ROOT.RooRealVar", weight:Optional[str]=None, n_bins:Optional[int]=None,
                  tree:Optional["ROOT.TTree"]=None, hist:Optional["ROOT.TH1"]=None):
    import ROOT
    ROOT.gROOT.SetBatch(True)
    
    if n_bins is None:
        n_bins = observable.numBins()
    vmin = observable.getMin()
    vmax = observable.getMax()
    name = observable.GetName()
    
    mask = 1*(tree is not None) + 2*(hist is not None)
    if mask not in [1, 2]:
        raise RuntimeError("either a tree or a histogram must be provided to setup model parameter templates")
    # case a tree is given
    if mask == 1:
        h_name = uuid.uuid4().hex
        hist = ROOT.TH1D(h_name, h_name, n_bins, vmin, vmax)
        if weight is not None:
            tree.Draw(f"{name} >> {h_name}", weight)
        else:
            tree.Draw(f"{name} >> {h_name}")
    # evaluate shape properties
    hist_max             = hist.GetMaximum()
    hist_bin_pos_max     = hist.GetMaximumBin()
    hist_pos_max         = hist.GetBinCenter(hist_bin_pos_max)
    hist_pos_FWHM_low    = hist.GetBinCenter(hist.FindFirstBinAbove(0.5 * hist_max))
    hist_pos_FWHM_high   = hist.GetBinCenter(hist.FindLastBinAbove(0.5 * hist_max))
    hist_sigma_effective = (hist_pos_FWHM_high - hist_pos_FWHM_low) / 2.355;
    # free memory
    if mask == 1:
        hist.Delete()
    parameters = {
        'muCBNom': ROOT.RooRealVar("muCBNom", "mean of CB", hist_pos_max, hist_pos_FWHM_low, hist_pos_FWHM_high),
        'sigmaCBNom': ROOT.RooRealVar("sigmaCBNom", "sigma of CB", hist_sigma_effective, 0., 5 * hist_sigma_effective),
        'alphaCBLo': ROOT.RooRealVar("alphaCBLo", "alpha of CB", 1, 0., 5.),
        'nCBLo': ROOT.RooRealVar("nCBLo", "n of CB", 10, 0., 200.),
        'alphaCBHi': ROOT.RooRealVar("alphaCBHi", "alpha of CB", 1, 0., 5.),
        'nCBHi': ROOT.RooRealVar("nCBHi", "n of CB", 10, 0., 200.)
    }
    return parameters

def template_ExpGaussExp(observable:"ROOT.RooRealVar", weight:Optional[str]=None,
                         tree:Optional["ROOT.TTree"]=None, hist:Optional["ROOT.TH1"]=None):
    parameters = {
        "EGE_mean": ROOT.RooRealVar("EGE_mean", "mean of EGE", 124.5, 123., 126.),
        "EGE_sigma": ROOT.RooRealVar("EGE_sigma", "sigma of EGE", 2.5, 0.5, 8.0),
        "EGE_kLo": ROOT.RooRealVar("EGE_kLo", "kLow of EGE", 2.5, 0.01, 10.0),
        "EGE_kHi": ROOT.RooRealVar("EGE_kHi", "kHigh of EGE", 2.4, 0.01, 10.0)
    }
    return parameters

def template_Bukin(observable:"ROOT.RooRealVar", weight:Optional[str]=None,
                   tree:Optional["ROOT.TTree"]=None, hist:Optional["ROOT.TH1"]=None):
    parameters = {
        "Bukin_Xp": ROOT.RooRealVar("Bukin_Xp", "peak position", 124.5, 123., 126.),
        "Bukin_sigp": ROOT.RooRealVar("Bukin_sigp", "peak width as FWHM divided by 2*sqrt(2*log(2))=2.35",
                                      0.5, 0.01, 2.0),
        "Bukin_xi": ROOT.RooRealVar("Bukin_xi", "peak asymmetry", 0.0, -1, 1),
        "Bukin_rho1": ROOT.RooRealVar("Bukin_rho1", "left tail", -0.1, -1.0, 0.0),
        "Bukin_rho2": ROOT.RooRealVar("Bukin_rho2", "right tail", 0.0, 0.0, 1.0)
    }
    return parameters

def template_Exp(observable:"ROOT.RooRealVar", weight:Optional[str]=None,
                 tree:Optional["ROOT.TTree"]=None, hist:Optional["ROOT.TH1"]=None):
    parameters = {
        "Exp_c": ROOT.RooRealVar("Exp_c", "Exp_c", 1, -10, 10)
    }
    return parameters

template_maps = {
    'DSCB': template_DSCB,
    'RooTwoSidedCBShape': template_DSCB,
    'ExpGaussExp': template_ExpGaussExp,
    'RooExpGaussExpShape': template_ExpGaussExp,
    'Bukin': template_Bukin,
    'RooBukinPdf': template_Bukin,
    'Exp': template_Exp,
    'RooExponential': template_Exp
}


def get_param_templates(model:str):
    if model not in template_maps:
        raise RuntimeError(f"no default parameter templates found for the model \"{model}\"")
    return template_maps[model]


import os
import sys
import time
from typing import Optional, Union, Dict, List, Any

from quickstats import semistaticmethod
from quickstats.components import AbstractObject
from quickstats.concurrent.logging import standard_log
from quickstats.utils.common_utils import execute_multi_tasks, is_valid_file

class AbstractRunner(AbstractObject):
    
    @property
    def config(self):
        return self._config
    
    def __init__(self, parallel:int=-1, timed:bool=True,
                 save_log:bool=True, cache:bool=True,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)

        self._config = {
            'cache': cache,
            'parallel': parallel,
            'save_log': save_log,            
            'timed': timed
        }

    def _prerun_batch(self):
        pass

    @semistaticmethod
    def _prerun_instance(self, **kwargs):
        pass
    
    @semistaticmethod
    def _run_instance(self, **kwargs):
        raise NotImplementedError
    
    @semistaticmethod
    def _cached_return(self, outname:str):
        raise NotImplementedError
        
    def run_instance(self, kwargs:Dict[str, Any]):
        self._prerun_instance(**kwargs)
        
        outname = kwargs.get("outname", None)
        
        if outname and (self.config['cache'] and os.path.exists(outname) and is_valid_file(outname)):
            self.stdout.info(f"INFO: Cached output from {outname}")
            return self._cached_return(outname)
        
        log_path = kwargs.get("log_path", None)
        if (outname and self.config['save_log']) and (not log_path):
            log_path = os.path.splitext(outname)[0] + ".log"
        
        if self.config['save_log']:
            with standard_log(log_path) as logger:
                return self._run_instance(**kwargs)
        else:
            return self._run_instance(**kwargs)
        
    def run_batch(self, argument_list, auxiliary_args:Optional[Dict]=None):
        parallel = self.config['parallel']
        t0 = time.time()
        self._prerun_batch()
        raw_result = execute_multi_tasks(self.run_instance, argument_list, parallel=parallel)
        results = self.postprocess(raw_result, auxiliary_args)
        t1 = time.time()
        dt = t1 - t0        
        if self.config['timed']:
            self.stdout.info('INFO: All jobs have finished. Total time taken: {:.3f} s'.format(dt))
        return results
    
    def postprocess(self, raw_result, auxiliary_args:Optional[Dict]=None):
        return raw_result


from typing import Optional, Union, List, Dict

class Systematic:
    def __init__(self):
        self.name = ""
        self.domain = ""
        self.process = ""
        self.whereto = ""
        self.constr_term = ""
        self.nominal = 0.
        self.beta = 0.
        self.errorlo = None
        self.errorhi = None
    def is_equal(self, other:"Systematic"):
        return (self.name == other.name) and (self.process == other.process) and (self.whereto == other.whereto)


from quickstats.interface.root.TObject import TObject

from quickstats.interface.root.macros import load_macros, load_macro

from quickstats.interface.root.TArrayData import TArrayData
from quickstats.interface.root.TH1 import TH1
from quickstats.interface.root.TH2 import TH2
from quickstats.interface.root.RooDataSet import RooDataSet
from quickstats.interface.root.RooAbsPdf import RooAbsPdf

from quickstats.interface.root.data_conversion import *

load_macros()


import os
import sys
import copy
import json
from typing import Optional, Union, Dict, List, Any
from itertools import repeat

import pandas as pd

import ROOT
from quickstats import semistaticmethod
from quickstats.parsers import ParamParser
from quickstats.concurrent import ParameterisedRunner
from quickstats.components import Likelihood
from quickstats.utils.common_utils import batch_makedirs
from quickstats.maths.numerics import pretty_value

class ParameterisedLikelihood(ParameterisedRunner):
    def __init__(self, input_file:str, param_expr:str,
                 filter_expr:Optional[str]=None, exclude_expr:Optional[str]=None,
                 data_name:str="combData", config:Optional[Dict]=None,
                 outdir:str="output", outname:str="{poi_names}.json",
                 cache:bool=True, save_log:bool=True, parallel:int=-1,
                 verbosity:Optional[Union[int, str]]="INFO"):
        
        super().__init__(file_expr=None, param_expr=param_expr,
                         filter_expr=filter_expr, exclude_expr=exclude_expr,
                         parallel=parallel, timed=True, save_log=save_log,
                         cache=cache, verbosity=verbosity)
        
        self.attributes = {
            'input_file': input_file,
            'data_name': data_name,
            'config': config,
            'outdir': outdir,
            'outname': outname
        }
        
        self.attributes['poi_name'] = ParamParser._get_param_str_attributes(param_expr)

    def _prerun_batch(self):
        outdir = self.attributes['outdir']
        cache_dir = self.get_cache_dir()
        batch_makedirs([outdir, cache_dir])
    
    @semistaticmethod
    def _prerun_instance(self, filename:str, mode:int, poi_val:Optional[Union[float, Dict[str, float]]]=None, **kwargs):
        if mode == 2:
            param_str = "("+ParamParser.val_encode_parameters(poi_val)+")"
            self.stdout.info(f"INFO: Evaluating conditional NLL {param_str} for the workspace {filename}")
        elif mode == 1:
            self.stdout.info(f"INFO: Evaluating unconditional NLL for the workspace {filename}")
    
    @semistaticmethod
    def _cached_return(self, outname:str):
        with open(outname, 'r') as f:
            likelihood = json.load(f)
            if 'uncond_fit' in likelihood:
                return likelihood['uncond_fit']['nll']
            elif 'cond_fit' in likelihood:
                return likelihood['cond_fit']['nll']
            else:
                raise RuntimeError("unexpected output (expect only conditional/unconditional fit in each task)")
    
    @semistaticmethod
    def _run_instance(self, filename:str, mode:int,
                      poi_name:Optional[Union[str, List[str]]]=None,
                      poi_val:Optional[Union[float, Dict[str, float]]]=None,
                      data_name:str="combData",
                      config:Optional[Dict]=None,
                      outname:Optional[str]=None,
                      **kwargs):
        try:
            if mode not in [1, 2]:
                error_msg = "only unconditional/conditional fit is allowed in parameterised likelihood runner"
                self.stdout.error(error_msg)
                raise RuntimeError(error_msg)
            if config is None:
                config = {}
            verbosity = config.pop("verbosity", "INFO")
            do_minos = config.pop("do_minos", False)
            likelihood = Likelihood(filename=filename, poi_name=poi_name, data_name=data_name, 
                                    config=config, verbosity=verbosity)
            fit_result = likelihood.nll_fit(poi_val, mode=mode, do_minos=do_minos)
            # save results
            if outname is not None:
                with open(outname, 'w') as outfile:
                    json.dump(fit_result, outfile)
                print('INFO: Saved NLL result to {}'.format(outname))
            if mode == 1:
                return fit_result['uncond_fit']['nll']
            elif mode == 2:
                return fit_result['cond_fit']['nll']
        except Exception as e:
            sys.stdout.write(f"{e}\n")
            return None
        
    def get_cache_dir(self):
        return os.path.join(self.attributes['outdir'], 'cache')
    
    def prepare_task_inputs(self):
        
        poi_names = self.attributes['poi_name']
        if len(poi_names) == 0:
            raise RuntimeError("no POI(s) to scan for")
            
        input_file = self.attributes['input_file']
        cache_dir  = self.get_cache_dir()
        outname    = "{param_str}.json"
        param_points = self.get_param_points(input_file)
        param_data = self.get_serialised_param_data(param_points, outdir=cache_dir, outname=outname)
        
        # None is for unconditional NLL
        poi_values = [None]
        # 1 is for unconditional NLL, 2 is for conditional NLL
        modes = [1] + [2]*len(param_points)
        
        for param_point in param_points:
            poi_values.append(param_point['internal_parameters'])
            
        outname_uncond = os.path.join(cache_dir, "{}_uncond.json".format("_".join(poi_names)))
        param_dpd_kwargs = {
            'poi_val': poi_values,
            'mode': modes,
            'outname': [outname_uncond] + param_data['outnames']
        }
        
        filename = list(set(param_data['filenames']))
        
        if len(filename) != 1:
            raise RuntimeError("multiple input files detected: {}".format(", ".join(filename)))
            
        param_ind_kwargs = {
            'filename': filename[0],
            'poi_name': self.attributes['poi_name'],
            'data_name': self.attributes['data_name'],
            'config': self.attributes['config']
        }
        
        self.set_param_ind_kwargs(**param_ind_kwargs)
        self.set_param_dpd_kwargs(**param_dpd_kwargs)
        kwarg_set = self.create_kwarg_set()
        auxiliary_args = {
            'points': poi_values
        }
        if not kwarg_set:
            raise RuntimeError("no parameter point to scan for")        
        return kwarg_set, auxiliary_args
    
    def postprocess(self, raw_result, auxiliary_args:Optional[Dict]=None):
        points = auxiliary_args['points']
        for nll, poi_values in zip(raw_result, points):
            if nll is None:
                if poi_values is None:
                    raise RuntimeError(f'NLL evaluation failed for the unconditional fit. '
                                       'Please check the log file for more details.')
                else:
                    param_str = parser.val_encode_parameters(poi_values)
                    raise RuntimeError(f'NLL evaluation failed for the conditional fit ({param_str}).'
                                       'Please check the log file for more details.')
        uncond_nll = raw_result[0]
        data = {'nll':[], 'qmu':[]}
        for nll in raw_result:
            data['nll'].append(nll)
            data['qmu'].append(2*(nll-uncond_nll))
        # only 1 POI
        if len(self.attributes['poi_name']) == 1:
            poi_name = self.attributes['poi_name'][0]
            data['mu'] = []
            for poi_values in points:
                mu = pretty_value(poi_values[poi_name]) if poi_values is not None else None
                data['mu'].append(mu)
        # multi POI case
        else:
            for poi_name in self.attributes['poi_name']:
                data[poi_name] = []
                for poi_values in points:
                    mu = pretty_value(poi_values[poi_name]) if poi_values is not None else None
                    data[poi_name].append(mu)
        outdir  = self.attributes['outdir']
        outname = self.attributes['outname'].format(poi_names="_".join(self.attributes['poi_name']))
        outpath = os.path.join(outdir, outname.format(poi_name=poi_name))
        with open(outpath, 'w') as outfile:
            json.dump(data, outfile, indent=3)


import os
import math
import time
import json
from itertools import repeat
from typing import Optional, Dict, List, Union

import numpy as np
import ROOT
from quickstats.components import ExtendedModel, ExtendedMinimizer
from quickstats.maths.numerics import approx_n_digit, str_encode_value, str_decode_value, pretty_value
from quickstats.utils import common_utils

def load_cache(fname:str):
    if (fname is not None) and os.path.exists(fname):
        with open(fname, 'r') as file:
            data = json.load(file)
        return data
    else:
        return None

from quickstats.components import AnalysisObject
from quickstats.utils.common_utils import parse_config

class Likelihood(AnalysisObject):
    
    def __init__(self, filename:str, poi_name:Optional[Union[str, List[str]]]=None,
                 data_name:str='combData', 
                 config:Optional[Dict]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        config = parse_config(config)
        config['filename']  = filename
        config['poi_name']  = poi_name
        config['data_name'] = data_name
        config['verbosity'] = verbosity
        self._inherit_init(super().__init__, **config)
        self.roofit_result = None
    
    @staticmethod
    def _parse_poi_val(pois:"ROOT.RooArgSet", poi_val:Optional[Union[Dict[str, float], float]]=None):
        if poi_val is None:
            poi_val = {}
            for poi in pois:
                poi_val[poi.GetName()] = poi.getVal()
        elif not isinstance(poi_val, dict):
            poi_val_tmp = poi_val
            poi_val = {}
            for poi in pois:
                poi_val[poi.GetName()] = poi_val_tmp
        else:
            # fill in missing values
            for poi in pois:
                poi_name = poi.GetName()
                if poi_name not in poi_val:
                    poi_val[poi_name] = poi.getVal()
        return poi_val
        
    def nll_fit(self, poi_val:Optional[Union[Dict[str, float], float]]=None,
                mode:int=2, snapshot_name:Optional[str]="currentSnapshot",
                do_minos:bool=False):
        if not isinstance(self.poi, ROOT.RooArgSet):
            pois = ROOT.RooArgSet(self.poi)
        else:
            pois = self.poi
        
        poi_val = self._parse_poi_val(pois, poi_val)

        if snapshot_name is None:
            self.save_snapshot("tmpSnapshot")
            snapshot_name = "tmpSnapshot"
        
        nll_uncond = None
        nll_cond   = None
        muhat      = None
        
        tmp_do_minos = self.minimizer.config['minos']
        
        self.minimizer.config['minos'] = do_minos
        if do_minos:
            self.minimizer.minos_set = pois
        
        start_time = time.time()
        roofit_result = {}
        muhat = {}
        muhat_errlo = {}
        muhat_errhi = {}
        
        # no poi fit
        if mode == 3:
            # restore initial parameter values after each fit
            self.load_snapshot(snapshot_name)
            self.minimizer.create_nll()
            prefit_nll = self.minimizer.nll.getVal()
            fit_status = self.minimizer.minimize()
            roofit_result = self.minimizer.fit_result
            self.save_snapshot("nllFit")
            postfit_nll = self.minimizer.nll.getVal()
            fit_time = time.time() - start_time
            
        # unconditional nll
        if mode in [0, 1]:
            # restore initial parameter values after each fit
            self.load_snapshot(snapshot_name)
            # fix other pois if they are not studied
            for poi in self.model.pois:
                poi.setConstant(1)
            for poi in pois:
                poi.setConstant(0)
                val = poi_val[poi.GetName()]
                poi.setVal(val)
            fit_status_uncond = self.minimizer.minimize()
            roofit_result['uncond'] = self.minimizer.fit_result
            self.save_snapshot("uncondFit")
            nll_uncond = self.minimizer.nll.getVal()
            for poi in pois:
                poi_name = poi.GetName()
                muhat[poi_name]       = poi.getVal()
                muhat_errlo[poi_name] = poi.getErrorLo()
                muhat_errhi[poi_name] = poi.getErrorHi()
        fit_time_uncond = time.time() - start_time
        # conditional nll
        if mode in [0, 2]:
            # restore initial parameter values after each fit
            self.load_snapshot(snapshot_name)
            # fix other pois if they are not studied
            for poi in self.model.pois:
                poi.setConstant(1)
            for poi in pois:
                poi.setConstant(1)
                val = poi_val.get(poi.GetName(), poi.getVal())
                poi.setVal(val)
            fit_status_cond = self.minimizer.minimize()
            roofit_result['cond'] = self.minimizer.fit_result
            self.save_snapshot("condFit")
            nll_cond = self.minimizer.nll.getVal()
        fit_time_cond = time.time() - fit_time_uncond - start_time
        
        fit_result = {}
        self.stdout.info("INFO: NLL evaluation completed with")
        if mode in [0, 1]:
            if len(pois) == 1:
                poi_name = pois.first().GetName()
                muhat = muhat[poi_name]
                muhat_errlo = muhat_errlo[poi_name]
                muhat_errhi = muhat_errhi[poi_name]
                self.stdout.info("mu hat = ".rjust(15)+f"{muhat:.5f}")
            else:
                best_fit_str = ", ".join([f"{name}={value:.5f}" for name, value in muhat.items()])
                self.stdout.info("best fit : ".rjust(15) + f"{best_fit_str}")
            self.stdout.info("uncond NLL = ".rjust(15) + f"{nll_uncond:.5f}")
            fit_result['uncond_fit'] = {
                'muhat': muhat,
                'muhat_errlo': muhat_errlo,
                'muhat_errhi': muhat_errhi,
                'nll': nll_uncond,
                'status': fit_status_uncond,
                'time': fit_time_uncond
            }
        if mode in [0, 2]:
            if len(pois) == 1:
                poi_name = pois.first().GetName()
                mu = pretty_value(poi_val[poi_name])
                self.stdout.info("mu = ".rjust(15)+f"{mu:.5f}")
            else:
                mu = poi_val
                mu_str = ", ".join([f"{name}={pretty_value(value)}" for name, value in poi_val.items()])
                self.stdout.info("mu : ".rjust(15) + f"{mu_str}")
            self.stdout.info("cond NLL = ".rjust(15)+f"{nll_cond:.5f}")
            fit_result['cond_fit'] = {
                'mu': mu,
                'nll': nll_cond,
                'status': fit_status_cond,
                'time': fit_time_cond
            }
            
        if mode == 0:
            pnll = nll_cond - nll_uncond
            qmu  = 2*pnll
            self.stdout.info("PNLL = ".rjust(15) + "{:.5f}".format(nll_cond - nll_uncond))
            fit_result['pnll'] = pnll
            fit_result['qmu']  = qmu
            
            if (qmu >= 0):
                ndof = len(pois)
                if ndof == 1:
                    # ndof = 1 case
                    poi_name = pois.first().GetName()
                    x0 = poi_val[poi_name]
                    significance = math.sqrt(qmu)
                    pvalue = 1 - ROOT.Math.normal_cdf(significance, 1, x0)
                else:
                    x0 = list(set(poi_val.values()))
                    if len(x0) > 1:
                        fit_result['significance'] = None
                        fit_result['pvalue'] = None
                    else:
                        pvalue = ROOT.Math.chisquared_cdf_c(qmu, ndof, x0[0])
                        significance = ROOT.RooStats.PValueToSignificance(pvalue)
                fit_result['significance'] = significance
                fit_result['pvalue'] = pvalue
            else:
                fit_result['significance'] = None
                fit_result['pvalue'] = None
                
        if mode != 3:
            self.stdout.info("time = ".rjust(15) + \
                             "(uncond_fit) {:.3f}, (cond_fit) {:.3f}".format(fit_time_uncond, fit_time_cond))
            fit_result['total_time'] = fit_time_uncond + fit_time_cond
        else:
            self.stdout.info("prefit NLL = ".rjust(18) + f"{prefit_nll:.5f}")
            self.stdout.info("postfit NLL = ".rjust(18) + f"{postfit_nll:.5f}")
            self.stdout.info("time = ".rjust(18) + f"{fit_time:.3f}")
            fit_result['prefit_nll'] = prefit_nll
            fit_result['postfit_nll'] = postfit_nll
            fit_result['total_time'] = fit_time
        
        self.roofit_result = roofit_result
        
        # finallizing
        self.minimizer.config['minos'] = tmp_do_minos
        self.minimizer.minos_set = ROOT.RooArgSet()
        
        return fit_result
        
    def evaluate_nll(self, poi_val:Optional[Union[Dict[str, float], float]]=None,
                     mode:int=2, snapshot_name:Optional[str]="currentSnapshot"):
        """Evalute post-fit NLL value
        
        Arguments:
            poi_val: (Optional) float
                Value of parameter of interest during fit. If not specified, the original value is used.
            mode: int
                Evaluation mode. Choose from:
                    0: Evaluate nll_mu - nll_muhat
                    1: Evaluate unconditional NLL (nll_muhat)
                    2: Evaluate conditional NLL (nll_mu)
                    3: Evaluate post-fit NLL (POI state remain as is)
            snapshot_name: (Optional) str
                Name of snapshot to load before fitting.
        """
        fit_result = self.nll_fit(poi_val=poi_val, mode=mode, snapshot_name=snapshot_name)
        
        if mode == 0:
            return fit_result['pnll']
        if mode == 1:
            return fit_result['uncond_fit']['nll']
        if mode == 2:
            return fit_result['cond_fit']['nll']
        if mode == 3:
            return fit_result['postfit_nll']


import ROOT
import time
import numpy as np

from quickstats.interface.cppyy.vectorize import np_type_str_maps

def array2root(array_data, fname:str, tree_name:str, multithread:bool=True):
    if (not ROOT.IsImplicitMTEnabled()) and (multithread):
        ROOT.EnableImplicitMT()
    elif (ROOT.IsImplicitMTEnabled()) and (not multithread):
        ROOT.DisableImplicitMT()
    columns = list(array_data.keys())
    snapshot_templates = []
    for column in columns:
        template_type = np_type_str_maps.get(array_data[column].dtype, None)
        if template_type is None:
            raise ValueError(f"unsupported array type \"{array_data[column].dtype}\"")
        snapshot_templates.append(template_type)
    snapshot_templates = tuple(snapshot_templates)
    df = ROOT.RDF.MakeNumpyDataFrame(array_data)
    df.Snapshot.__getitem__(snapshot_templates)(tree_name, fname, columns)


from typing import List, Optional

from quickstats.utils.root_utils import declare_expression

ROOT_MACROS = \
{
    "TH1Utils":
    """
    namespace TH1Utils {
        template<typename T>
        std::vector<T> GetBinErrorArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinError(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinErrorUpArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinErrorUp(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinErrorLowArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinErrorLow(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinCenterArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinCenter(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinContentArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinContent(bin_index));
            return result;
        }
      
        template<typename T>
        std::vector<T> GetBinWidthArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinWidth(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinLowEdgeArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinLowEdge(bin_index));
            return result;
        }    
    }; 
    """,
    "TAxisUtils": 
    """
    namespace TAxisUtils{
        template<typename T>
        std::vector<T> GetBinLowEdgeArray(TAxis* ax, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = ax->GetNbins();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(ax->GetBinLowEdge(bin_index));
            return result;
        }
        template<typename T>
        std::vector<T> GetBinCenterArray(TAxis* ax, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = ax->GetNbins();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(ax->GetBinCenter(bin_index));
            return result;
        }
        template<typename T>
        std::vector<T> GetBinWidthArray(TAxis* ax, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = ax->GetNbins();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(ax->GetBinWidth(bin_index));
            return result;
        }
    };
    """,    
    "THistUtils": 
    """
    namespace THistUtils{
        template<typename T>
        std::vector<double> GetPoissonError(const std::vector<T> data){
            Double_t ym, yp;
            auto inst = RooHistError::instance();
            const int data_size = data.size();
            std::vector<double> result(2*data_size);
            for (size_t i = 0; i < data_size; i++){
                inst.getPoissonInterval(Int_t(data[i] + 0.5), ym, yp, 1);
                result[i] = data[i] - ym;
                result[data_size + i] = yp - data[i];
            }
            return result;
        }
    };
    """,
    "RFUtils":
    """
    namespace RFUtils{
        struct DatasetStruct {
          std::vector<double> observable_values;
          std::vector<double> weights;
          std::vector<std::string> category_labels;
          std::vector<int> category_index;
          DatasetStruct(const size_t &n_entries, const size_t &n_obs){
              observable_values = std::vector<double>(n_obs * n_entries);
              weights = std::vector<double>(n_entries);
              category_labels = std::vector<std::string>(n_entries);
              category_index = std::vector<int>(n_entries);
          }
        } ;
        
        DatasetStruct ExtractCategoryData(const RooAbsData* dataset, const RooArgSet* observables, const RooCategory* cat){
            const size_t n_entries = dataset->numEntries();
            const size_t n_obs = observables->size();
            DatasetStruct result(n_entries, n_obs);
            TIterator* iter(observables->createIterator());
            RooRealVar* obs;
            for (size_t i = 0; i < n_entries; i++){
                dataset->get(i);
                for (size_t j = 0; j < n_obs; j++)
                    result.observable_values[i + j * n_entries] = ((RooRealVar*)(*observables)[j])->getVal();
                result.weights[i] = dataset->weight();
                result.category_labels[i] = cat->getLabel();
                result.category_index[i] = cat->getIndex();
            }
            return result;
        }        
    };
    """
}

def load_macro(macro_name:str):
    expression = ROOT_MACROS.get(macro_name, None)
    if expression is None:
        raise ValueError(f"`{macro_name}` is not a built-in quickstats macro."
                         " Available macros are: {}".format(",".join(list(ROOT_MACROS))))
    declare_expression(expression, macro_name)

def load_macros(macro_names:Optional[List[str]]=None):
    if macro_names is None:
        macro_names = list(ROOT_MACROS)
    for macro_name in macro_names:
        load_macro(macro_name)


from typing import Dict, Union, List, Optional

import numpy as np

import ROOT

from quickstats import semistaticmethod
from quickstats.interface.cppyy.vectorize import as_np_array

class RooAbsPdf:
    @staticmethod
    def extract_sum_pdfs_by_category(pdf:ROOT.RooAbsPdf, poi:Optional[ROOT.RooRealVar]=None):
        pdf_class = pdf.ClassName()
        if pdf_class != "RooSimultaneous":
            raise RuntimeError(f"input pdf must be a RooSimultaneous instance (`{pdf_class}` received)")
        cat = pdf.indexCat()
        n_cat = cat.size()
        result = {}
        for i in range(n_cat):
            cat.setBin(i)
            cat_name = cat.getLabel()
            cat_pdf = pdf.getPdf(cat_name)
            cat_pdf_class = cat_pdf.ClassName()
            if cat_pdf_class != "RooProdPdf":
                raise RuntimeError(f"category pdf must be a RooProdPdf instance (`{cat_pdf_class}` received)")
            target_pdf = [i for i in cat_pdf.pdfList() if i.ClassName() == "RooRealSumPdf" and i != cat_pdf]
            if not target_pdf:
                raise RuntimeError(f"category pdf does not contain a RooRealSumPdf component")
            if len(target_pdf) > 1:
                raise RuntimeError(f"expect only one RooRealSumPdf component from category pdf but {len(target_pdf)} found")
            target_pdf = target_pdf[0]
            if poi is None:
                result[cat_name] = [i for i in pdf.getComponents()]
            else:
                result[cat_name] = [i for i in pdf.getComponents() if i.dependsOn(poi)]
        return result


import uuid
import array
import numpy as np

import ROOT

from quickstats.interface.root import TObject, TArrayData
from quickstats.interface.root import load_macro
from quickstats.interface.cppyy.vectorize import as_vector, as_np_array, as_c_array, np_type_str_maps

class TH1(TObject):
    
    FUNDAMENTAL_TYPE = ROOT.TH1
    
    def __init__(self, h:ROOT.TH1, dtype:str='double', 
                 underflow_bin:int=0,
                 overflow_bin:int=0):
        self.dtype = dtype
        self.underflow_bin = underflow_bin
        self.overflow_bin  = overflow_bin
        self.init(h)
        
    def init(self, h):
        self.bin_content  = self.GetBinContentArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
        self.bin_error    = self.GetBinErrorArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
        self.bin_center   = self.GetBinCenterArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
        self.bin_width    = self.GetBinWidthArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
        self.bin_low_edge = self.GetBinLowEdgeArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
        
    @staticmethod
    def _to_ROOT(name:str, bin_content, bin_width=None, bin_center=None, bin_low_edge=None,
                 bin_error=None, title:str=None):
        if ((bin_low_edge is None) and (bin_center is None)) or \
           ((bin_low_edge is not None) and (bin_center is not None)):
            raise ValueError("must provide either bin low edge or bin center values")
        n_bins = len(bin_content)
        if bin_center is not None:
            if bin_width is None:
                if n_bins < 2:
                    raise ValueError("unable to determine bin width with only one bin")
                bin_width = (bin_center[1] - bin_center[0])
            bin_low_edge = bin_center - bin_width/2
        if len(bin_low_edge) == n_bins:
            bin_width = (bin_low_edge[-1] - bin_low_edge[-2])
            bin_low_edge = np.insert(bin_low_edge, len(bin_low_edge), bin_low_edge[-1] + bin_width)
        if title is None:
            title = name
        dtype = bin_content.dtype
        if np_type_str_maps[dtype] == "float":
            h_func = ROOT.TH1F
        else:
            h_func = ROOT.TH1D
        h = h_func(name, title, n_bins, as_c_array(bin_low_edge))
        if bin_error is None:
            bin_error = np.zeros((n_bins,))
        for i in range(n_bins):
            h.SetBinContent(i + 1, bin_content[i])
            h.SetBinError(i + 1, bin_error[i])
        return h
    
    def to_ROOT(self, name:str, title:str=None):
        if title is None:
            title = name
        return self._to_ROOT(name, self.bin_content, bin_low_edge=self.bin_low_edge,
                             bin_error=self.bin_error, title=title)
        
    @staticmethod
    def GetBinContentArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinContentArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)        
        
    @staticmethod
    def GetBinErrorArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinErrorArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)
                     
    @staticmethod
    def GetBinErrorUpArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinErrorUpArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)

    @staticmethod
    def GetBinErrorLowArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinErrorLowArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)                     

    @staticmethod
    def GetBinCenterArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinCenterArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)
    
    @staticmethod
    def GetBinWidthArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinWidthArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)

    @staticmethod
    def GetBinLowEdgeArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        c_vector = ROOT.TH1Utils.GetBinLowEdgeArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)
    
    @staticmethod
    def GetPoissonError(data:np.ndarray):
        c_data = as_vector(data)
        c_result  = ROOT.THistUtils.GetPoissonError(c_data)
        result = as_np_array(c_result)
        data_size = data.shape[0]
        pois_error = {
            'lo': result[:data_size],
            'hi': result[data_size:]
        }
        return pois_error
        


import os
import sys
import copy
import json
from typing import Optional, Union, Dict, List, Any
from itertools import repeat

import pandas as pd

from quickstats import semistaticmethod
from quickstats.parsers import ParamParser
from quickstats.concurrent import ParameterisedRunner
from quickstats.components import AsymptoticCLs
from quickstats.utils.common_utils import batch_makedirs

class ParameterisedAsymptoticCLs(ParameterisedRunner):
    def __init__(self, input_path:str, config:Optional[Dict]=None,
                 file_expr:Optional[str]=None, param_expr:Optional[str]=None,
                 filter_expr:Optional[str]=None, exclude_expr:Optional[str]=None,
                 outdir:str="output", outname:str="limits.json",
                 save_summary:bool=True, cache:bool=True,
                 save_log:bool=True, parallel:int=-1,
                 verbosity:Optional[Union[int, str]]="INFO"):
        
        super().__init__(file_expr=file_expr, param_expr=param_expr,
                         filter_expr=filter_expr, exclude_expr=exclude_expr,
                         parallel=parallel, timed=True, save_log=save_log,
                         cache=cache, verbosity=verbosity)
        
        self.attributes = {
            'input_path': input_path,
            'config': config,
            'outdir': outdir,
            'outname': outname,
            'save_summary': save_summary
        }

    def _prerun_batch(self):
        outdir = self.attributes['outdir']
        cache_dir = self.get_cache_dir()
        batch_makedirs([outdir, cache_dir])
    
    @semistaticmethod
    def _prerun_instance(self, filename:str, parameters:Optional[Dict[str, Any]]=None, **kwargs):
        if parameters:
            param_str = "("+ParamParser.val_encode_parameters(parameters)+")"
        else:
            param_str = ""
        self.stdout.info(f"INFO: Evaluating limit for the workspace {filename} {param_str}")
    
    @semistaticmethod
    def _cached_return(self, outname:str):
        with open(outname, 'r') as f:
            limits = json.load(f)
            return limits
    
    @semistaticmethod
    def _run_instance(self, filename:str, config:Dict[str, Any],
                      outname:Optional[str]=None,
                      save_summary:bool=True, **kwargs):
        try:
            config['filename'] = filename
            asymptotic_cls = AsymptoticCLs(**config)
            asymptotic_cls.evaluate_limits()
            if outname is not None:
                asymptotic_cls.save(outname, summary=save_summary)
            return asymptotic_cls.limits
        except Exception as e:
            sys.stdout.write(f"{e}\n")
            return {}
        
    def get_cache_dir(self):
        return os.path.join(self.attributes['outdir'], 'cache')
        
    def prepare_task_inputs(self):
        input_path = self.attributes['input_path']
        cache_dir  = self.get_cache_dir()
        outname    = "{param_str}.json"
        param_points = self.get_param_points(input_path)
        param_data = self.get_serialised_param_data(param_points, outdir=cache_dir, outname=outname)
        
        configs = []
        base_config = self.attributes['config']
        if base_config is None:
            base_config = {}
        for param_point in param_points:
            new_config = copy.deepcopy(base_config)
            int_params = param_point['internal_parameters']
            val_expr = ParamParser.val_encode_parameters(int_params)
            base_fix_param = new_config.get("fix_param", None)
            fix_param = []
            if base_fix_param:
                fix_param.append(base_fix_param)
            if val_expr:
                fix_param.append(val_expr)
            new_config['fix_param'] = ",".join(fix_param)
            configs.append(new_config)
        param_dpd_kwargs = {
            'filename': param_data['filenames'],
            'config': configs,
            'outname': param_data['outnames'],
            'parameters': param_data['parameters']
        }
        param_ind_kwargs = {
            'save_summary': self.attributes['save_summary']
        }
        self.set_param_ind_kwargs(**param_ind_kwargs)
        self.set_param_dpd_kwargs(**param_dpd_kwargs)
        kwarg_set = self.create_kwarg_set()
        auxiliary_args = {
            'filenames': param_data['filenames'],
            'parameters': param_data['parameters']
        }
        if not kwarg_set:
            raise RuntimeError("no parameter point to scan for")
        return kwarg_set, auxiliary_args
    
    def postprocess(self, raw_result, auxiliary_args:Optional[Dict]=None):
        filenames = auxiliary_args["filenames"]
        parameters = auxiliary_args["parameters"]
        
        final_result = []
        for filename, params, limit in zip(filenames, parameters, raw_result):
            if len(limit) == 0:
                if params:
                    param_str = "("+ParamParser.val_encode_parameters(params)+")"
                else:
                    param_str = ""                
                raise RuntimeError(f'Job failed for the input "{filename}" {param_str}. '
                                   'Please check the log file for more details.')
            final_result.append({**params, **limit})
        final_result = pd.DataFrame(final_result).to_dict('list')
        
        outdir     = self.attributes['outdir']
        outname    = self.attributes['outname']
        
        if outname is not None:
            outpath = os.path.join(outdir, outname)
            with open(outpath, "w") as f:
                json.dump(final_result, f, indent=2)


from typing import Optional, Union, Dict, List

from cycler import cycler

import numpy as np

from quickstats.plots.color_schemes import QUICKSTATS_PALETTES

from quickstats.plots import AbstractPlot
from quickstats.plots.template import suggest_markersize, ratio_frames, centralize_axis, create_transform
from quickstats.utils.common_utils import combine_dict

class PdfDistributionPlot(AbstractPlot):
    
    STYLES = {
        'errorbar': {
            "marker": 'o',
            "markersize": None,
            "linewidth": 0,
            "elinewidth": 1,
            "capsize": 3,
            "capthick": 1
        }
    }
    
    CONFIG = {
        "blind_linestyle": "--",
        'ratio_line_styles':{
            'color': 'gray',
            'linestyle': '--'
        }        
    }
    
    def __init__(self, collective_data, blind_range:Optional[List[float]]=None,
                 plot_options:Optional[Dict]=None,
                 label_map:Optional[Dict]=None,
                 color_cycle:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Union[Dict, str]]=None,
                 figure_index:Optional[int]=None,
                 config:Optional[Dict]=None):
        super().__init__(color_cycle=color_cycle,
                         styles=styles, figure_index=figure_index,
                         analysis_label_options=analysis_label_options,
                         config=config)
        self.collective_data = collective_data
        self.legend_order = list(collective_data.keys())
        if plot_options is None:
            self.plot_options = {}
        else:
            self.plot_options = plot_options
        if label_map is None:
            self.label_map = {}
        else:
            self.label_map = label_map

        self.colors = {}

    def get_error_data(self, data):
        xerr = data.get('xerr', None)
        yerr = data.get('yerr', None)
        xerrlo = data.get('xerrlo', None)
        yerrlo = data.get('yerrlo', None)
        xerrhi = data.get('xerrhi', None)
        yerrhi = data.get('yerrhi', None)
        if all(err is None for err in [xerr, yerr, xerrlo, yerrlo, xerrhi, yerrhi]):
            return {"xerr": None, "yerr": None}
        if (xerr is not None) and ((xerrlo is not None) or (xerrhi is not None)):
            raise ValueError("invalid data: received both \"xerr\" and \"xerrlo\"/\"xerrhi\" attributes")
        if (yerr is not None) and ((yerrlo is not None) or (yerrhi is not None)):
            raise ValueError("invalid data: received both \"yerr\" and \"yerrlo\"/\"yerrhi\" attributes")
        error_data = {}
        if xerr is not None:
            error_data['xerr'] = xerr
        elif ((xerrlo is not None) or (xerrhi is not None)):
            error_data['xerr'] = (xerrlo, xerrhi)
        else:
            error_data['xerr'] = None
        if yerr is not None:
            error_data['yerr'] = yerr
        elif ((yerrlo is not None) or (yerrhi is not None)):
            error_data['yerr'] = (yerrlo, yerrhi)
        else:
            error_data['yerr'] = None
        return error_data
    
    def _get_selected_err(self, err, mask):     
        if err is None:
            return None
        elif isinstance(err, tuple):
            new_err = []
            for err_i in err:
                new_err.append(err_i[mask])
            return tuple(new_err)
        else:
            return err[mask]
        
    def draw_single_data(self, ax, data, label:str, blind_range:Optional[List[float]]=None,
                         plot_options:Optional[Dict]=None, show_errorbar:bool=True):
        pdata = self.process_data(data, blind_range=blind_range)
        has_errorbar = not ((pdata['xerr'] is None) and (pdata['yerr'] is None))
        draw_blind = blind_range is not None
        if not has_errorbar:
            combined_options = combine_dict(self.styles['plot'], plot_options)
            combined_options['label'] = label
            if draw_blind:
                # draw sideband low
                handle_sblo = ax.plot(pdata['x'][pdata['sideband_lo']], 
                                      pdata['y'][pdata['sideband_lo']],
                                      **combined_options)
                # avoid going through internal color cycle for the three regions
                combined_options['color'] = handle_sblo[0].get_color()
                # draw sideband high
                handle_sbhi = ax.plot(pdata['x'][pdata['sideband_hi']], 
                                      pdata['y'][pdata['sideband_hi']],
                                      **combined_options)
                combined_options['linestyle'] = self.config['blind_linestyle']
                combined_options['label'] = f"{label} (blind)"
                handle_blind = ax.plot(pdata['x'][pdata['blind']], 
                                       pdata['y'][pdata['blind']],
                                       **combined_options)
                handle = [handle_sblo[0], handle_blind[0]]
            else:
                handle = ax.plot(pdata['x'], pdata['y'], **combined_options)
                handle = handle[0]
        else:
            combined_options = combine_dict(self.styles['errorbar'], plot_options)
            combined_options['label'] = label
            if combined_options.get('markersize', None) is None:
                nbins = len(pdata['x'])
                combined_options['markersize'] = suggest_markersize(nbins)
            if not show_errorbar:
                combined_options['elinewidth'] = 0
                combined_options['capsize'] = 0
                combined_options['capthick'] = 0
            if draw_blind:
                xerr = self._get_selected_err(pdata['xerr'], pdata['sideband'])
                yerr = self._get_selected_err(pdata['yerr'], pdata['sideband'])
                handle = ax.errorbar(pdata['x'][pdata['sideband']],
                                     pdata['y'][pdata['sideband']],
                                     xerr=xerr,
                                     yerr=yerr,
                                     **combined_options)
            else:
                handle = ax.errorbar(pdata['x'], pdata['y'],
                                     xerr=pdata['xerr'],
                                     yerr=pdata['yerr'],
                                      **combined_options)
        return handle
    
    def process_data(self, data, blind_range:Optional[List[float]]=None):
        processed_data = {}
        processed_data['x'] = data['x']
        processed_data['y'] = data['y']
        error_data = self.get_error_data(data)
        processed_data['xerr'] = error_data['xerr']
        processed_data['yerr'] = error_data['yerr']
        if blind_range is not None:
            x = processed_data['x']
            blind_min = blind_range[0]
            blind_max = blind_range[1]
            sideband_lo = (x <= blind_min)
            sideband_hi = (x >= blind_max)
            blind = (x > blind_min) & (x < blind_max)
            processed_data['sideband_lo'] = sideband_lo
            processed_data['sideband_hi'] = sideband_hi
            processed_data['blind'] = blind
            processed_data['sideband'] = ~blind
        return processed_data
    
    def draw_comparison(self, ax, reference_data, target_data, mode:str="ratio",
                        plot_options:Optional[Dict]=None, show_errorbar:bool=True, 
                        blind_range:Optional[List[float]]=None,
                        draw_ratio_line:bool=True):
        pdata_ref = self.process_data(reference_data, blind_range=blind_range)
        pdata_tgt = self.process_data(target_data, blind_range=blind_range)
        if not np.allclose(pdata_tgt['x'], pdata_ref['x']):
            raise RuntimeError("cannot compare two distributions with different binnings")
        pdata_comp = {}
        pdata_comp['x'] = pdata_ref['x']
        if pdata_tgt['yerr'] is not None:
            raise RuntimeError("target distribution must be a pdf (i.e. no error information)")
        pdata_comp['yerr'] = pdata_ref['yerr']
        pdata_comp['xerr'] = pdata_ref['xerr']
        if mode == "ratio":
            vmode = 0
        elif mode in ["diff", "difference"]:
            vmode = 1
        else:
            raise ValueError(f"unsupported mode \"{mode}\", choose between \"ratio\" and \"difference\"")
        if vmode == 0:
            pdata_comp['y'] = pdata_ref['y'] / pdata_tgt['y']
            if pdata_ref['yerr'] is not None:
                if isinstance(pdata_ref['yerr'], tuple):
                    pdata_comp['yerr'] = (pdata_ref['yerr'][0] / pdata_tgt['y'], 
                                          pdata_ref['yerr'][1] / pdata_tgt['y'])
                else:
                    pdata_comp['yerr'] = pdata_ref['yerr'] / pdata_tgt['y']
        elif vmode == 1:
            pdata_comp['y'] = pdata_ref['y'] / pdata_tgt['y']
        combined_options = combine_dict(self.styles['errorbar'], plot_options)
        if combined_options.get('markersize', None) is None:
            nbins = len(pdata_comp['x'])
            combined_options['markersize'] = suggest_markersize(nbins)
        if not show_errorbar:
            combined_options['elinewidth'] = 0
            combined_options['capsize'] = 0
            combined_options['capthick'] = 0
        draw_blind = blind_range is not None
        if draw_blind:
            xerr = self._get_selected_err(pdata_comp['xerr'], pdata_ref['sideband'])
            yerr = self._get_selected_err(pdata_comp['yerr'], pdata_ref['sideband'])
            handle = ax.errorbar(pdata_comp['x'][pdata_ref['sideband']],
                                 pdata_comp['y'][pdata_ref['sideband']],
                                 xerr=xerr,
                                 yerr=yerr,
                                 **combined_options)
        else:
            handle = ax.errorbar(pdata_comp['x'], pdata_comp['y'],
                                 xerr=pdata_comp['xerr'],
                                 yerr=pdata_comp['yerr'],
                                  **combined_options)
        if vmode == 0:
            centralize_axis(ax, which="y", ref_value=1)
            if draw_ratio_line:
                ratio_linestyles = self.config['ratio_line_styles']
                ax.axhline(1, zorder=0, **ratio_linestyles)
        return handle
        
    
    def draw(self, xlabel:str="", ylabel:str="",
             logx:bool=False, logy:bool=False,
             blind_range:Optional[List[float]]=None,
             show_errorbar:bool=True, 
             comparison_options:Optional[str]=None):
        
        if comparison_options:
            ax, ax_ratio = self.draw_frame(ratio_frames, logx=logx, logy=logy)
        else:
            ax = self.draw_frame(logx=logx, logy=logy)
        
        for name in self.collective_data:
            data = self.collective_data[name]
            label = self.label_map.get(name, name)
            plot_options = self.plot_options.get(name, {})
            handle = self.draw_single_data(ax, data, label,
                                           blind_range=blind_range,
                                           plot_options=plot_options,
                                           show_errorbar=show_errorbar)
            # case draw blind
            if isinstance(handle, list) and len(handle) == 2:
                blind_name = f"{name}_blind"
                self.legend_order.append(blind_name)
                self.update_legend_handles({name: handle[0], blind_name: handle[1]})
                self.colors[name] = handle[0].get_color()
            else:
                self.update_legend_handles({name: handle})
                if not handle.get_children():
                    self.colors[name] = handle.get_color()
                else:
                    self.colors[name] = handle[0].get_color()
        
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        
        if comparison_options:
            reference = comparison_options['reference']
            target = comparison_options['target']
            mode = comparison_options.get('mode', "ratio")
            label = comparison_options.get('label', "Data / Fit")
            draw_ratio_line = comparison_options.get('draw_ratio_line', True)
            plot_options = self.plot_options.get(reference, {})
            plot_options['color'] = self.colors[reference]
            self.draw_comparison(ax_ratio,
                                 self.collective_data[reference],
                                 self.collective_data[target],
                                 mode=mode,
                                 plot_options=plot_options,
                                 show_errorbar=show_errorbar,
                                 blind_range=blind_range,
                                 draw_ratio_line=draw_ratio_line)
            self.draw_axis_components(ax_ratio, xlabel=ax.get_xlabel(), ylabel=label)
            ax.set(xlabel=None)
            ax.tick_params(axis="x", labelbottom=False)
        return ax


import os
import click

@click.command(name='likelihood_scan')
@click.option('-i', '--input_file', required=True, 
              help='Path to the input workspace file.')
@click.option('-p', '--param_expr', default=None,
              help='\b\n Parameter expression, e.g.'
                   '\b\n 1D scan: "poi_name=<poi_min>_<poi_max>_<step>"'
                   '\b\n 2D scan: "poi_1_name=<poi_1_min>_<poi_1_max>_<step_1>,'
                   '\b\n           poi_2_name=<poi_2_min>_<poi_2_max>_<step_2>"')
@click.option('--filter', 'filter_expr', default=None, show_default=True,
              help='\b Filter parameter points by expression.\n'
                   '\b Example: "mass=2*,350,400,450;klambda=1.*,2.*,-1.*,-2.*"\n'
                   '\b Refer to documentation for more information\n')
@click.option('--exclude', 'exclude_expr', default=None, show_default=True,
              help='\b Exclude parameter points by expression.\n'
                   '\b Example: "mass=2*,350,400,450;klambda=1.*,2.*,-1.*,-2.*"\n'
                   '\b Refer to documentation for more information\n')
@click.option('--min', 'poi_min', type=float, default=None, 
              help='(deprecated) Minimum POI value to scan.')
@click.option('--max', 'poi_max', type=float, default=None, 
              help='(deprecated) Maximum POI value to scan.')
@click.option('--step', 'poi_step', type=float, default=None, 
              help='(deprecated) Scan interval.')
@click.option('-p', '--poi', 'poi_name', default=None, show_default=True,
              help='(deprecated) POI to scan. If not specified, the first POI from the workspace is used.')
@click.option('--cache/--no-cache', default=True, show_default=True,
              help='Cache existing result')
@click.option('-o', '--outname', default='{poi_names}.json', show_default=True,
              help='Name of output.')
@click.option('--outdir', default='likelihood_scan', show_default=True,
              help='Output directory.')
@click.option('--save_log/--skip_log', default=True, show_default=True,
              help='Save log file')
@click.option('-w', '--workspace', 'ws_name', default=None, show_default=True,
              help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, show_default=True,
              help='Name of model config. Auto-detect by default.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of dataset.')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, show_default=True,
              help='Name of initial snapshot')
@click.option('-r', '--profile', 'profile_param', default="", show_default=True,
              help='Parameters to profile')
@click.option('-f', '--fix', 'fix_param', default="", show_default=True,
              help='Parameters to fix')
@click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, show_default=True,
              help='Use constrained NLL (i.e. include systematics)')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('--binned/--unbinned', 'binned_likelihood', default=True, show_default=True,
              help='Binned likelihood')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('-v', '--verbosity', default='INFO', show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
              help='Verbosity level.')
@click.option('--parallel', type=int, default=-1, show_default=True,
              help='\b\n Parallelize job across the N workers.'
                   '\b\n Case  0: Jobs are run sequentially (for debugging).'
                   '\b\n Case -1: Jobs are run across N_CPU workers.')
def likelihood_scan(**kwargs):
    """
    Evaluate a set of parmeterised likelihood values
    """
    _kwargs = {}
    for arg_name in ["input_file", "param_expr", "data_name", "outdir", "filter_expr",
                     "exclude_expr", "outname", "cache", "save_log", "parallel", "verbosity"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _kwargs['config'] = kwargs
    
    # for backward compatibility
    _deprecated_kwargs = {}
    for arg_name in ["poi_min", "poi_max", "poi_step", "poi_name"]:
        _deprecated_kwargs[arg_name] = kwargs.pop(arg_name)
    if all(_deprecated_kwargs[arg_name] is not None for arg_name in ["poi_min", "poi_max", "poi_step"]):
        if _kwargs['param_expr'] is not None:
            raise RuntimeError("either `param_expr` or (`poi_min`, `poi_max`, `poi_step`) should be "
                               "given for 1D likelihood scan")
        print("WARNING: Likelihood scan using `poi_min`, `poi_max` and `poi_step` is "
              "deprecated. Use --param_expr \"<poi_name>=<poi_min>_<poi_max>_<poi_step>\" instead.")
        from quickstats.components import AnalysisBase
        analysis = AnalysisBase(_kwargs['input_file'], poi_name=_deprecated_kwargs['poi_name'],
                                data_name=_kwargs['data_name'], verbosity="WARNING")
        poi_name = analysis.poi.GetName()
        poi_min = _deprecated_kwargs['poi_min']
        poi_max = _deprecated_kwargs['poi_max']
        poi_step = _deprecated_kwargs['poi_step']
        _kwargs['param_expr'] = f"{poi_name}={poi_min}_{poi_max}_{poi_step}"
    elif (not all(_deprecated_kwargs[arg_name] is None for arg_name in ["poi_min", "poi_max", "poi_step"])) or \
         (_kwargs['param_expr'] is None):
        raise RuntimeError("either `param_expr` or (`poi_min`, `poi_max`, `poi_step`) should be "
                           "given for 1D likelihood scan")
    from quickstats.concurrent import ParameterisedLikelihood
    runner = ParameterisedLikelihood(**_kwargs)
    runner.run()



BUILTIN_METHODS = \
{
    "Namespace_ROOT_VecOps": "using namespace ROOT::VecOps;", 
    "Namespace_ROOT_Math": "using namespace ROOT::Math;",
    "SizeOf" : "template<typename T>\n"
               "auto SizeOf(T vec)\n"
               "{return vec.size();}",
    "CosineThetaStar": "template<typename T1, typename T2>\n"
                       "auto CosineThetaStar(T1 P1, T2 P2){\n"
                       "auto P12 = P1 + P2;\n"
                       "double P1_plus  = (P1.E() + P1.Pz())/1.41421356;\n"
                       "double P2_plus  = (P2.E() + P2.Pz())/1.41421356;\n"
                       "double P1_minus = (P1.E() - P1.Pz())/1.41421356;\n"
                       "double P2_minus = (P2.E() - P2.Pz())/1.41421356;\n"
                       "double M12_squared = pow(P12.M(),2);\n"
                       "double PT12_squared = pow(P12.Pt(),2);\n"
                       "return 2*((P1_plus*P2_minus -P1_minus*P2_plus)/"
                       "sqrt(M12_squared*(M12_squared+PT12_squared)))*(P12.Pz()/abs(P12.Pz()));}\n",
    "LVUtils": "LorentzVector<PtEtaPhiM4D<double>> getLV_PtEtaPhiM(const float &pt, const float &eta, const float &phi, const float &M){\n"
               "    LorentzVector<PtEtaPhiM4D<double>> result{pt, eta, phi, M};\n"
               "    return result;\n"
               "}\n"
               "LorentzVector<PtEtaPhiE4D<double>> getLV_PtEtaPhiE(const float &pt, const float &eta, const float &phi, const float &E){\n"
               "    LorentzVector<PtEtaPhiE4D<double>> result{pt, eta, phi, E};\n"
               "    return result;\n"
               "}\n"
               "RVec<LorentzVector<PtEtaPhiM4D<double>>> getLVArray_PtEtaPhiM(const RVec<float> &pt, const RVec<float> &eta, \n"
               "                                                             const RVec<float> &phi, const RVec<float> &M){\n"
               "    RVec<LorentzVector<PtEtaPhiM4D<double>>> result;\n"
               "    for (size_t i=0; i<pt.size(); i++){\n"
               "        LorentzVector<PtEtaPhiM4D<double>> lv{pt[i], eta[i], phi[i], M[i]};\n"
               "        result.push_back(lv);\n"
               "    }\n"
               "    return result;\n"
               "}\n"
               "RVec<LorentzVector<PtEtaPhiM4D<double>>> getLVArray_PtEtaPhiM(const RVec<float> &pt, const RVec<float> &eta, \n"
               "                                                             const RVec<float> &phi, const float &M){\n"
               "    RVec<LorentzVector<PtEtaPhiM4D<double>>> result;\n"
               "    for (size_t i=0; i<pt.size(); i++){\n"
               "        LorentzVector<PtEtaPhiM4D<double>> lv{pt[i], eta[i], phi[i], M};\n"
               "        result.push_back(lv);\n"
               "    }\n"
               "    return result;\n"
               "}\n"
               "RVec<LorentzVector<PtEtaPhiE4D<double>>> getLVArray_PtEtaPhiE(const RVec<float> &pt, const RVec<float> &eta, \n"
               "                                                             const RVec<float> &phi, const RVec<float> &E){\n"
               "    RVec<LorentzVector<PtEtaPhiE4D<double>>> result;\n"
               "    for (size_t i=0; i<pt.size(); i++){\n"
               "        LorentzVector<PtEtaPhiE4D<double>> lv{pt[i], eta[i], phi[i], E[i]};\n"
               "        result.push_back(lv);\n"
               "    }\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "LorentzVector<PtEtaPhiE4D<double>> getLVSum_PtEtaPhiE(const RVec<LorentzVector<T>> &lv_vec, int n=0){\n"
               "    LorentzVector<PtEtaPhiE4D<double>> result;\n"
               "    if ((n==0)||(n>lv_vec.size()))\n"
               "        n = (int)lv_vec.size();\n"
               "    for (size_t i=0; i<n; i++)\n"
               "        result += lv_vec[i];\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "LorentzVector<PtEtaPhiM4D<double>> getLVSum_PtEtaPhiM(const RVec<LorentzVector<T>> &lv_vec, int n=0){\n"
               "    LorentzVector<PtEtaPhiM4D<double>> result;\n"
               "    if ((n==0)||(n>lv_vec.size()))\n"
               "        n = (int)lv_vec.size();\n"
               "    for (size_t i=0; i<n; i++)\n"
               "        result += lv_vec[i];\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "RVec<float> getArrayPt(const RVec<LorentzVector<T>> &lv_vec){\n"
               "    RVec<float> result;\n"
               "    for (auto &lv: lv_vec)\n"
               "        result.push_back(lv.pt());\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "RVec<float> getArrayEta(const RVec<LorentzVector<T>> &lv_vec){\n"
               "    RVec<float> result;\n"
               "    for (auto &lv: lv_vec)\n"
               "        result.push_back(lv.eta());\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "RVec<float> getArrayPhi(const RVec<LorentzVector<T>> &lv_vec){\n"
               "    RVec<float> result;\n"
               "    for (auto &lv: lv_vec)\n"
               "        result.push_back(lv.phi());\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "RVec<float> getArrayE(const RVec<LorentzVector<T>> &lv_vec){\n"
               "    RVec<float> result;\n"
               "    for (auto &lv: lv_vec)\n"
               "        result.push_back(lv.E());\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "RVec<float> getArrayM(const RVec<LorentzVector<T>> &lv_vec){\n"
               "    RVec<float> result;\n"
               "    for (auto &lv: lv_vec)\n"
               "        result.push_back(lv.M());\n"
               "    return result;\n"
               "}\n"
               "template<typename T>\n"
               "float getPt(const LorentzVector<T> &lv_vec){\n"
               "    return lv_vec.pt();\n"
               "}\n"
               "template<typename T>\n"
               "float getEta(const LorentzVector<T> &lv_vec){\n"
               "    return lv_vec.eta();\n"
               "}\n"
               "template<typename T>\n"
               "float getPhi(const LorentzVector<T> &lv_vec){\n"
               "    return lv_vec.phi();\n"
               "}\n"
               "template<typename T>\n"
               "float getM(const LorentzVector<T> &lv_vec){\n"
               "    return lv_vec.M();\n"
               "}\n"
               "template<typename T>\n"
               "float getE(const LorentzVector<T> &lv_vec){\n"
               "    return lv_vec.E();\n"
               "}\n"
               "template<typename T>\n"
               "std::vector<T> RVec2Vec(const RVec<T> v){\n"
               "    return std::vector<T>{v.begin(), v.end()};\n"
               "}\n"
}


from .abstract_runner import AbstractRunner
from .parameterised_runner import ParameterisedRunner
from .parameterised_asymptotic_cls import ParameterisedAsymptoticCLs
from .parameterised_likelihood import ParameterisedLikelihood